/* THIS CODE CARRIES NO GUARANTEE OF USABILITY OR FITNESS FOR ANY PURPOSE.
 * WHILE THE AUTHORS HAVE TRIED TO ENSURE THE PROGRAM WORKS CORRECTLY,
 * IT IS STRICTLY USE AT YOUR OWN RISK.  */

#include "rgbe.h"
#include <cmath>
#include <cstring>
#include <cctype>
#include <iostream>
#include "def.h"
using namespace std;

/* This file contains code to read and write four byte rgbe file format
 developed by Greg Ward.  It handles the conversions between rgbe and
 pixels consisting of floats.  The data is assumed to be an array of floats.
 By default there are three floats per pixel in the order red, green, blue.
 (RGBE_DATA_??? values control this.)  Only the mimimal header reading and 
 writing is implemented.  Each routine does error checking and will return
 a status value as defined below.  This code is intended as a skeleton so
 feel free to modify it to suit your needs.

 (Place notice here if you modified the code.)
 posted to http://www.graphics.cornell.edu/~bjw/
 written by Bruce Walter  (bjw@graphics.cornell.edu)  5/26/95
 based on code written by Greg Ward
*/

/* offsets to red, green, and blue components in a data (float) pixel */
enum RGBE_DATA
{
 RGBE_DATA_RED = 0,
 RGBE_DATA_GREEN = 1,
 RGBE_DATA_BLUE = 2,
// number of floats per pixel 
 RGBE_DATA_SIZE = 3
};

enum rgbe_error_codes {
  rgbe_read_error,
  rgbe_write_error,
  rgbe_format_error,
  rgbe_memory_error
};

/* default error routine.  change this to change error handling */
static bool rgbe_error(int rgbe_error_code, char *msg)
{
  switch (rgbe_error_code) {
  case rgbe_read_error:
    cerr << "RGBE read error" << endl;
    break;
  case rgbe_write_error:
    cerr << "RGBE write error" << endl;
    break;
  case rgbe_format_error:
    cerr << "RGBE bad file format: " << msg << endl;
    break;
  default:
  case rgbe_memory_error:
    cerr << "RGBE error: " << msg << endl;
  }
  return false;
}

/* standard conversion from float pixels to rgbe pixels */
/* note: you can remove the "inline"s if your compiler complains about it */
static inline void 
float2rgbe(unsigned char rgbe[4], const color &c)
{
  float v;
  int e;

  v = c.red;
  if (c.green > v) v = c.green;
  if (c.blue > v) v = c.blue;
  if (v < 1e-32) {
    rgbe[0] = rgbe[1] = rgbe[2] = rgbe[3] = 0;
  }
  else {
    v = frexpf(v,&e) * 256.0f/v;
    rgbe[0] = (unsigned char) (c.red * v);
    rgbe[1] = (unsigned char) (c.green * v);
    rgbe[2] = (unsigned char) (c.blue * v);
    rgbe[3] = (unsigned char) (e + 128);
  }
}

/* standard conversion from rgbe to float pixels */
/* note: Ward uses ldexp(col+0.5,exp-(128+8)).  However we wanted pixels */
/*       in the range [0,1] to map back into the range [0,1].            */
static inline void 
rgbe2float(color *c, unsigned char rgbe[4])
{
  float f;

  if (rgbe[3]) {   /*nonzero pixel*/
    f = ldexpf(1.0f,rgbe[3]-(int)(128+8));
    c->red = rgbe[0] * f;
    c->green = rgbe[1] * f;
    c->blue = rgbe[2] * f;
  }
  else
    c->red = c->green = c->blue = 0.0;
}

/* minimal header reading.  modify if you want to parse more information */
bool RGBE_ReadHeader(ifstream &currentfile, int *width, int *height, rgbe_header_info *info)
{
    char buf[128];
    int found_format;
    float tempf;
    int i;
    found_format = 0;
    if (info) {
        info->valid = 0;
        info->programtype[0] = 0;
        info->gamma = info->exposure = 1.0;
    }
    currentfile.getline(buf, sizeof(buf)/sizeof(buf[0]));
    if (!currentfile)
        return rgbe_error(rgbe_read_error,NULL);
    if ((buf[0] != '#')||(buf[1] != '?')) {
        /* if you want to require the magic token then uncomment the next line */
        /*return rgbe_error(rgbe_format_error,"bad initial token"); */
    }
    else if (info) {
        info->valid |= RGBE_VALID_PROGRAMTYPE;
        for( i=0; i < int(sizeof(info->programtype)) - 1; i++) {
            if ((buf[i+2] == 0) || isspace(buf[i+2]))
                break;
            info->programtype[i] = buf[i+2];
        }
        info->programtype[i] = 0;
        currentfile.getline(buf, sizeof(buf)/sizeof(buf[0]));
        if (!currentfile)
            return rgbe_error(rgbe_read_error,NULL);
    }
    for(;;) {
        if ((buf[0] == 0)||(buf[0] == '\n'))
            return rgbe_error(rgbe_format_error,"no FORMAT specifier found");
        else if (strcmp(buf,"FORMAT=32-bit_rle_rgbe") == 0)
            break;       /* format found so break out of loop */
        currentfile.getline(buf, sizeof(buf)/sizeof(buf[0]));
        if (!currentfile)
            return rgbe_error(rgbe_read_error,NULL);
    }
    currentfile.getline(buf, sizeof(buf)/sizeof(buf[0]));
    if (!currentfile)
        return rgbe_error(rgbe_read_error,NULL);
    if (info && (sscanf(buf,"GAMMA=%g",&tempf) == 1)) {
        info->gamma = tempf;
        info->valid |= RGBE_VALID_GAMMA;
    }
    else if (info && (sscanf(buf,"EXPOSURE=%g",&tempf) == 1)) {
        info->exposure = tempf;
        info->valid |= RGBE_VALID_EXPOSURE;
    }

    currentfile.getline(buf, sizeof(buf)/sizeof(buf[0]));
    if (!currentfile)
        return rgbe_error(rgbe_read_error,NULL);
    if (buf[0] != '\0')
        return rgbe_error(rgbe_format_error,
        "missing blank line after FORMAT specifier");
    currentfile.getline(buf, sizeof(buf)/sizeof(buf[0]));
    if (!currentfile)
        return rgbe_error(rgbe_read_error,NULL);
    if (sscanf(buf,"-Y %d +X %d",height,width) < 2)
        return rgbe_error(rgbe_format_error,"missing image size specifier");
    return true;
}

/* simple read routine.  will not correctly handle run length encoding */
bool RGBE_ReadPixels(ifstream &currentfile, color *data, int numpixels)
{
  unsigned char rgbe[4];

  while(numpixels-- > 0) {
    currentfile.read((char*)rgbe, sizeof(rgbe));
    if (!currentfile)
      return rgbe_error(rgbe_read_error,NULL);
    rgbe2float(data, rgbe);
    data ++;
  }
  return true;
}

bool RGBE_ReadPixels_RLE(ifstream &currentfile, color *data, int scanline_width,
                         int num_scanlines)
{
    unsigned char rgbe[4], *scanline_buffer, *ptr, *ptr_end;
    int i, count;
    unsigned char buf[2];

    if ((scanline_width < 8)||(scanline_width > 0x7fff))
        /* run length encoding is not allowed so read flat*/
        return RGBE_ReadPixels(currentfile,data,scanline_width*num_scanlines);
    scanline_buffer = NULL;
    /* read in each successive scanline */
    while(num_scanlines > 0) {
        currentfile.read((char*)rgbe,sizeof(rgbe));
        if (!currentfile) {
            free(scanline_buffer);
            return rgbe_error(rgbe_read_error,NULL);
        }
        if ((rgbe[0] != 2)||(rgbe[1] != 2)||(rgbe[2] & 0x80)) {
            /* this file is not run length encoded */
            rgbe2float(data,rgbe);
            data += RGBE_DATA_SIZE;
            free(scanline_buffer);
            return RGBE_ReadPixels(currentfile,data,scanline_width*num_scanlines-1);
        }
        if ((((int)rgbe[2])<<8 | rgbe[3]) != scanline_width) {
            free(scanline_buffer);
            return rgbe_error(rgbe_format_error,"wrong scanline width");
        }
        if (scanline_buffer == NULL)
            scanline_buffer = (unsigned char *)
            malloc(sizeof(unsigned char)*4*scanline_width);
        if (scanline_buffer == NULL) 
            return rgbe_error(rgbe_memory_error,"unable to allocate buffer space");

        ptr = &scanline_buffer[0];
        /* read each of the four channels for the scanline into the buffer */
        for(i=0;i<4;i++) {
            ptr_end = &scanline_buffer[(i+1)*scanline_width];
            while(ptr < ptr_end) {
                currentfile.read((char*)buf,sizeof(buf[0])*2);
                if (!currentfile) {
                    free(scanline_buffer);
                    return rgbe_error(rgbe_read_error,NULL);
                }
                if (buf[0] > 128) {
                    /* a run of the same value */
                    count = buf[0]-128;
                    if ((count == 0)||(count > ptr_end - ptr)) {
                        free(scanline_buffer);
                        return rgbe_error(rgbe_format_error,"bad scanline data");
                    }
                    while(count-- > 0)
                        *ptr++ = buf[1];
                }
                else {
                    /* a non-run */
                    count = buf[0];
                    if ((count == 0)||(count > ptr_end - ptr)) {
                        free(scanline_buffer);
                        return rgbe_error(rgbe_format_error,"bad scanline data");
                    }
                    *ptr++ = buf[1];
                    if (--count > 0) {
                        currentfile.read((char*)ptr,sizeof(*ptr)*count);
                        if (!currentfile) {
                            free(scanline_buffer);
                            return rgbe_error(rgbe_read_error,NULL);
                        }
                        ptr += count;
                    }
                }
            }
        }
        /* now convert data from buffer into floats */
        for(i=0;i<scanline_width;i++) {
            rgbe[0] = scanline_buffer[i];
            rgbe[1] = scanline_buffer[i+scanline_width];
            rgbe[2] = scanline_buffer[i+2*scanline_width];
            rgbe[3] = scanline_buffer[i+3*scanline_width];
            rgbe2float(data ,rgbe);
            data ++;
        }
        num_scanlines--;
    }
    free(scanline_buffer);
    return true;
}


